{-
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
-}
module ErrorMonad where {

import Util;
import List;

returna :: forall a e. a -> Eithera e a;
returna a = Righta a;

errora :: forall e a. e -> Eithera e a;
errora a = Lefta a;

checkB :: forall e. Bool -> e -> Eithera e ();
checkB b e = (if b then returna () else errora e);

isValue :: forall e a. Eithera e a -> Bool;
isValue p = (case p of {
              Lefta l -> False;
              Righta r -> True;
            });

bind :: forall e a b. Eithera e a -> (a -> Eithera e b) -> Eithera e b;
bind p vq = (case p of {
              Lefta a -> Lefta a;
              Righta a -> vq a;
            });

catcha :: forall e a f. Eithera e a -> (e -> Eithera f a) -> Eithera f a;
catcha p eq = (case p of {
                Lefta a -> eq a;
                Righta a -> Righta a;
              });

foldR :: forall a b. (a -> [a] -> b -> b) -> [a] -> b -> b;
foldR f (x : xs) b = f x xs (foldR f xs b);
foldR f [] b = b;

checkF :: forall a b. (a -> Bool) -> (a -> b) -> a -> Eithera b ();
checkF f g x = (if f x then returna () else errora (g x));

checkAllMI :: forall a e. (a -> Eithera e ()) -> [a] -> Eithera (a, e) ();
checkAllMI f xs =
  foldra (\ x iq -> bind (catcha (f x) (\ xa -> errora (x, xa))) (\ xa -> iq))
    xs (returna ());

checkSubset :: forall a. (Eq a) => [a] -> [a] -> Eithera a ();
checkSubset xs ys =
  catcha (checkAllMI (checkF (\ x -> member x ys) id) xs)
    (\ x -> errora (snd x));

checkDisjoint :: forall a. (Eq a) => [a] -> [a] -> Eithera a ();
checkDisjoint xs ys =
  catcha (checkAllMI (checkF (\ x -> not (member x ys)) id) xs)
    (\ x -> errora (snd x));

checkAllCombinations ::
  forall a b. (a -> a -> Eithera b ()) -> [a] -> Eithera ((a, a), b) ();
checkAllCombinations c xs =
  foldR (\ x ys res ->
          catcha
            (checkAllMI
              (\ y ->
                bind (catcha (c x y) (\ xa -> errora ((x, y), xa)))
                  (\ xa -> res))
              ys)
            (\ xa -> errora (snd xa)))
    xs (returna ());

checkPartition :: forall a. (Eq a) => [[a]] -> Eithera a ();
checkPartition cs =
  catcha (checkAllCombinations checkDisjoint cs) (\ x -> errora (snd x));

}
