{-
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
-}
module Nat where {

import qualified Product_Type;
import qualified HOL;

data Nat = Suc Nat.Nat | Zero_nat;

one_nat :: Nat.Nat;
one_nat = Nat.Suc Nat.Zero_nat;

plus_nat :: Nat.Nat -> Nat.Nat -> Nat.Nat;
plus_nat (Nat.Suc m) n = Nat.plus_nat m (Nat.Suc n);
plus_nat Nat.Zero_nat n = n;

nat_case :: forall t. t -> (Nat.Nat -> t) -> Nat.Nat -> t;
nat_case f1 f2 Nat.Zero_nat = f1;
nat_case f1 f2 (Nat.Suc nat) = f2 nat;

option_case :: forall t a. t -> (a -> t) -> Maybe a -> t;
option_case f1 f2 Nothing = f1;
option_case f1 f2 (Just a) = f2 a;

minus_nat :: Nat.Nat -> Nat.Nat -> Nat.Nat;
minus_nat (Nat.Suc m) (Nat.Suc n) = Nat.minus_nat m n;
minus_nat Nat.Zero_nat n = Nat.Zero_nat;
minus_nat m Nat.Zero_nat = m;

eq_nat :: Nat.Nat -> Nat.Nat -> Bool;
eq_nat Nat.Zero_nat Nat.Zero_nat = True;
eq_nat (Nat.Suc m) (Nat.Suc n) = Nat.eq_nat m n;
eq_nat Nat.Zero_nat (Nat.Suc a) = False;
eq_nat (Nat.Suc a) Nat.Zero_nat = False;

instance Eq Nat.Nat where {
  (==) = Nat.eq_nat;
};

times_nat :: Nat.Nat -> Nat.Nat -> Nat.Nat;
times_nat (Nat.Suc m) n = Nat.plus_nat n (Nat.times_nat m n);
times_nat Nat.Zero_nat n = Nat.Zero_nat;

the :: forall a. Maybe a -> a;
the (Just x) = x;

less_eq_nat :: Nat.Nat -> Nat.Nat -> Bool;
less_eq_nat (Nat.Suc m) n = Nat.less_nat m n;
less_eq_nat Nat.Zero_nat n = True;

less_nat :: Nat.Nat -> Nat.Nat -> Bool;
less_nat m (Nat.Suc n) = Nat.less_eq_nat m n;
less_nat n Nat.Zero_nat = False;

divmod :: Nat.Nat -> Nat.Nat -> (Nat.Nat, Nat.Nat);
divmod m n =
  (if Nat.eq_nat n Nat.Zero_nat || Nat.less_nat m n then (Nat.Zero_nat, m)
    else let {
           a = Nat.divmod (Nat.minus_nat m n) n;
           (q, aa) = a;
         } in (Nat.Suc q, aa));

mod_nat :: Nat.Nat -> Nat.Nat -> Nat.Nat;
mod_nat m n = snd (Nat.divmod m n);

div_nat :: Nat.Nat -> Nat.Nat -> Nat.Nat;
div_nat m n = fst (Nat.divmod m n);

}
