{-
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
-}
module Polynomial where {

import qualified Nat;
import qualified List;
import qualified Term;
import qualified Show;
import qualified Product_Type;
import qualified Util;
import qualified ErrorMonad;
import qualified HOL;
import qualified RedPair;

mul :: Nat.Nat -> Util.Cint -> Util.Cint;
mul n (Util.Pos m) = Util.Pos (Nat.times_nat n m);
mul n (Util.Neg m) = Util.Neg (Nat.times_nat n m);

data Poly a = Poly [(Nat.Nat, a)] Util.Cint;

data RedPair a = NegPolo [(a, (Util.Cint, [Nat.Nat]))]
  | Polo [(a, (Nat.Nat, [Nat.Nat]))];

addI :: Util.Cint -> Util.Cint -> Util.Cint;
addI (Util.Pos x) (Util.Neg y) =
  (if Nat.less_eq_nat y x then Util.Pos (Nat.minus_nat x y)
    else Util.Neg (Nat.minus_nat y x));
addI (Util.Neg x) (Util.Pos y) =
  (if Nat.less_eq_nat x y then Util.Pos (Nat.minus_nat y x)
    else Util.Neg (Nat.minus_nat x y));
addI (Util.Pos x) (Util.Pos y) = Util.Pos (Nat.plus_nat x y);
addI (Util.Neg x) (Util.Neg y) = Util.Neg (Nat.plus_nat x y);

ones :: Nat.Nat -> [Nat.Nat];
ones (Nat.Suc k) = Nat.one_nat : Polynomial.ones k;
ones Nat.Zero_nat = [];

poly_case ::
  forall v t. ([(Nat.Nat, v)] -> Util.Cint -> t) -> Polynomial.Poly v -> t;
poly_case f1 (Polynomial.Poly list cint) = f1 list cint;

polyMult :: forall a. Nat.Nat -> Polynomial.Poly a -> Polynomial.Poly a;
polyMult n (Polynomial.Poly vp c) =
  Polynomial.Poly (List.mapa (\ cx -> (Nat.times_nat n (fst cx), snd cx)) vp)
    (Polynomial.mul n c);

lookupVar ::
  forall a. (Eq a) => a -> [(Nat.Nat, a)] -> Maybe (Nat.Nat, [(Nat.Nat, a)]);
lookupVar x ((c, y) : vp) =
  (if HOL.eqop x y then Just (c, vp)
    else (case Polynomial.lookupVar x vp of {
           Nothing -> Nothing;
           Just a -> let {
                       (d, vpp) = a;
                     } in Just (d, (c, y) : vpp);
         }));
lookupVar x [] = Nothing;

sumPolyVars ::
  forall a. (Eq a) => [(Nat.Nat, a)] -> [(Nat.Nat, a)] -> [(Nat.Nat, a)];
sumPolyVars ((c, x) : vp1) vp2 =
  (case Polynomial.lookupVar x vp2 of {
    Nothing -> (c, x) : Polynomial.sumPolyVars vp1 vp2;
    Just a ->
      let {
        (d, vpp2) = a;
      } in (Nat.plus_nat c d, x) : Polynomial.sumPolyVars vp1 vpp2;
  });
sumPolyVars [] vp2 = vp2;

cleanPolyVars :: forall a. [(Nat.Nat, a)] -> [(Nat.Nat, a)];
cleanPolyVars ((c, x) : vp) =
  List.append (if Nat.eq_nat c Nat.Zero_nat then [] else [(c, x)])
    (Polynomial.cleanPolyVars vp);
cleanPolyVars [] = [];

polySum ::
  forall a.
    (Eq a) => Polynomial.Poly a -> Polynomial.Poly a -> Polynomial.Poly a;
polySum (Polynomial.Poly vp1 c) (Polynomial.Poly vp2 d) =
  Polynomial.Poly (Polynomial.cleanPolyVars (Polynomial.sumPolyVars vp1 vp2))
    (Polynomial.addI c d);

pleft ::
  forall a b.
    (Eq b) => (a -> Nat.Nat -> ([Nat.Nat], Util.Cint)) ->
                Term.Term a b -> Polynomial.Poly b;
pleft alpha (Term.Fun f ts) =
  let {
    a = alpha f (List.size_list ts);
    (list, c) = a;
    (Polynomial.Poly lista cint) =
      Polynomial.polySum (Polynomial.Poly [] c)
        (List.foldra Polynomial.polySum
          (List.mapa
            (\ ct ->
              Polynomial.polyMult (fst ct) (Polynomial.pleft alpha (snd ct)))
            (List.zipa list ts))
          (Polynomial.Poly [] (Util.Pos Nat.Zero_nat)));
  } in (case lista of {
         [] -> (case cint of {
                 Util.Neg n -> Polynomial.Poly [] (Util.Pos Nat.Zero_nat);
                 Util.Pos nat -> Polynomial.Poly [] (Util.Pos nat);
               });
         aa : listaa -> Polynomial.Poly (aa : listaa) cint;
       });
pleft alpha (Term.Var x) =
  Polynomial.Poly [(Nat.one_nat, x)] (Util.Pos Nat.Zero_nat);

pright ::
  forall a b.
    (Eq b) => (a -> Nat.Nat -> ([Nat.Nat], Util.Cint)) ->
                Term.Term a b -> Polynomial.Poly b;
pright alpha (Term.Fun f ts) =
  let {
    a = alpha f (List.size_list ts);
    (list, c) = a;
    (Polynomial.Poly vp aa) =
      Polynomial.polySum
        (List.foldra Polynomial.polySum
          (List.mapa
            (\ ct ->
              Polynomial.polyMult (fst ct) (Polynomial.pright alpha (snd ct)))
            (List.zipa list ts))
          (Polynomial.Poly [] (Util.Pos Nat.Zero_nat)))
        (Polynomial.Poly [] c);
  } in (case aa of {
         Util.Neg x -> Polynomial.Poly vp (Util.Pos Nat.Zero_nat);
         Util.Pos n -> Polynomial.Poly vp (Util.Pos n);
       });
pright alpha (Term.Var x) =
  Polynomial.Poly [(Nat.one_nat, x)] (Util.Pos Nat.Zero_nat);

genIndices :: forall a. Nat.Nat -> Integer -> [(Term.Term a [Char])];
genIndices (Nat.Suc n) i =
  Term.Var (List.append "x_" (Show.sshow_int i)) :
    Polynomial.genIndices n (i + 1);
genIndices Nat.Zero_nat uu = [];

genTerm :: forall a. a -> Nat.Nat -> Term.Term a [Char];
genTerm f n = Term.Fun f (Polynomial.genIndices n 0);

checkVPLists ::
  forall a. (Eq a) => [(Nat.Nat, a)] -> [(Nat.Nat, a)] -> Util.Eithera a ();
checkVPLists vp ((n, x) : vq) =
  (if Nat.eq_nat n Nat.Zero_nat then Polynomial.checkVPLists vp vq
    else (case Polynomial.lookupVar x vp of {
           Nothing -> ErrorMonad.errora x;
           Just a ->
             let {
               (m, vpp) = a;
             } in ErrorMonad.bind (ErrorMonad.checkB (Nat.less_eq_nat n m) x)
                    (\ xa -> Polynomial.checkVPLists vpp vq);
         }));
checkVPLists uu [] = ErrorMonad.returna ();

checkPoly ::
  forall a.
    (Eq a,
      Show.Sshow a) => Polynomial.Poly a ->
                         Polynomial.Poly a -> Util.Eithera [[Char]] ();
checkPoly (Polynomial.Poly vp c) (Polynomial.Poly vq d) =
  ErrorMonad.bind
    (ErrorMonad.checkB
      (case c of {
        Util.Neg x -> False;
        Util.Pos cn ->
          (case d of {
            Util.Neg x -> False;
            Util.Pos dn -> Nat.less_eq_nat dn cn;
          });
      })
      ["the constant of the left polynomial is too small"])
    (\ x ->
      ErrorMonad.catcha (Polynomial.checkVPLists vp vq)
        (\ xa ->
          ErrorMonad.errora
            [List.append (List.append "variable " (Show.sshow xa))
               " is problematic"]));

show_pair :: forall b. (Show.Sshow b) => (Nat.Nat, b) -> [Char];
show_pair (Nat.Suc (Nat.Suc n), v) =
  List.append (List.append (Show.sshow_nat (Nat.Suc (Nat.Suc n))) "*")
    (Show.sshow v);
show_pair (Nat.Suc Nat.Zero_nat, v) = Show.sshow v;
show_pair (Nat.Zero_nat, v) = [];

merge_entries :: [[Char]] -> [Char];
merge_entries (x : xs) =
  List.append (List.append " + " x) (Polynomial.merge_entries xs);
merge_entries [] = [];

show_number :: Util.Cint -> [Char];
show_number (Util.Pos (Nat.Suc n)) = Show.sshow_nat (Nat.Suc n);
show_number (Util.Neg (Nat.Suc n)) =
  List.append "-" (Show.sshow_nat (Nat.Suc n));
show_number (Util.Pos Nat.Zero_nat) = [];
show_number (Util.Neg Nat.Zero_nat) = [];

show_entries :: forall b. (Show.Sshow b) => Polynomial.Poly b -> [[Char]];
show_entries (Polynomial.Poly vp c) =
  List.filtera (\ entry -> not (entry == []))
    (Polynomial.show_number c : List.mapa Polynomial.show_pair vp);

sshow_poly :: forall b. (Show.Sshow b) => Polynomial.Poly b -> [Char];
sshow_poly p =
  (case Polynomial.show_entries p of {
    [] -> "0";
    d : es -> List.append d (Polynomial.merge_entries es);
  });

addAtLeastOne :: forall a. Polynomial.Poly a -> Polynomial.Poly a;
addAtLeastOne (Polynomial.Poly vp (Util.Neg uu)) =
  Polynomial.Poly vp (Util.Pos (Nat.Suc Nat.Zero_nat));
addAtLeastOne (Polynomial.Poly vp (Util.Pos cn)) =
  Polynomial.Poly vp (Util.Pos (Nat.Suc cn));

checkPolyGt ::
  forall v.
    (Eq v,
      Show.Sshow v) => Polynomial.Poly v ->
                         Polynomial.Poly v -> Util.Eithera [[Char]] ();
checkPolyGt p q =
  ErrorMonad.catcha (Polynomial.checkPoly p (Polynomial.addAtLeastOne q))
    (\ x ->
      ErrorMonad.errora
        (List.concata
           ["Could not orient ", Polynomial.sshow_poly p, " > ",
             Polynomial.sshow_poly q]
          : x));

checkSPolo ::
  forall f v.
    (Show.Sshow f, Eq v,
      Show.Sshow v) => (f -> Nat.Nat -> ([Nat.Nat], Util.Cint)) ->
                         (Term.Term f v, Term.Term f v) ->
                           Util.Eithera [[Char]] ();
checkSPolo alpha lr =
  ErrorMonad.catcha
    (Polynomial.checkPolyGt (Polynomial.pleft alpha (fst lr))
      (Polynomial.pright alpha (snd lr)))
    (\ x ->
      ErrorMonad.errora
        (List.concata
           ["Could not orient ", Term.sshow_term (fst lr), " > ",
             Term.sshow_term (snd lr)]
          : x));

checkPolyGe ::
  forall v.
    (Eq v,
      Show.Sshow v) => Polynomial.Poly v ->
                         Polynomial.Poly v -> Util.Eithera [[Char]] ();
checkPolyGe p q =
  ErrorMonad.catcha (Polynomial.checkPoly p q)
    (\ x ->
      ErrorMonad.errora
        (List.concata
           ["Could not orient ", Polynomial.sshow_poly p, " >= ",
             Polynomial.sshow_poly q]
          : x));

checkNSPolo ::
  forall f v.
    (Show.Sshow f, Eq v,
      Show.Sshow v) => (f -> Nat.Nat -> ([Nat.Nat], Util.Cint)) ->
                         (Term.Term f v, Term.Term f v) ->
                           Util.Eithera [[Char]] ();
checkNSPolo alpha lr =
  ErrorMonad.catcha
    (Polynomial.checkPolyGe (Polynomial.pleft alpha (fst lr))
      (Polynomial.pright alpha (snd lr)))
    (\ x ->
      ErrorMonad.errora
        (List.concata
           ["Could not orient ", Term.sshow_term (fst lr), " >= ",
             Term.sshow_term (snd lr)]
          : x));

checkListPositive :: [Nat.Nat] -> Util.Eithera [[Char]] ();
checkListPositive (n : ns) =
  ErrorMonad.bind
    (ErrorMonad.checkB (Nat.less_nat Nat.Zero_nat n)
      ["Coefficient 0 - not monotone"])
    (\ x -> Polynomial.checkListPositive ns);
checkListPositive [] = ErrorMonad.returna ();

checkPoloMono ::
  forall a. [(a, (Nat.Nat, [Nat.Nat]))] -> Util.Eithera [[Char]] ();
checkPoloMono ((uu, (uv, coeffs)) : alpha) =
  ErrorMonad.bind (Polynomial.checkListPositive coeffs)
    (\ x -> Polynomial.checkPoloMono alpha);
checkPoloMono [] = ErrorMonad.returna ();

showInterpretation ::
  forall a. (a -> Nat.Nat -> ([Nat.Nat], Util.Cint)) -> a -> Nat.Nat -> [Char];
showInterpretation alpha f n =
  Polynomial.sshow_poly (Polynomial.pleft alpha (Polynomial.genTerm f n));

createRepresentation ::
  forall a.
    (Eq a,
      Show.Sshow a) => (a -> Nat.Nat -> ([Nat.Nat], Util.Cint)) ->
                         [(a, Nat.Nat)] -> [[Char]];
createRepresentation alpha fs =
  "polynomial interpretation: " :
    List.mapa
      (\ (a @ (g, n)) ->
        List.concata
          ["Pol(", Show.sshow g, "/", Show.sshow_nat n, ") = ",
            Polynomial.showInterpretation alpha g n])
      (List.remdups fs);

redPairToPrePolyInter ::
  forall a. Polynomial.RedPair a -> [(a, (Util.Cint, [Nat.Nat]))];
redPairToPrePolyInter (Polynomial.NegPolo inter) = inter;
redPairToPrePolyInter (Polynomial.Polo inter) =
  List.mapa
    (\ (a @ (f, aa)) ->
      let {
        (i, coeffs) = aa;
      } in (f, (Util.Pos i, coeffs)))
    inter;

createPolyInter ::
  forall a.
    (Eq a) => [(a, (Util.Cint, [Nat.Nat]))] ->
                a -> Nat.Nat -> ([Nat.Nat], Util.Cint);
createPolyInter ((g, (i, coeffs)) : rest) f n =
  (if HOL.eqop f g && Nat.eq_nat n (List.size_list coeffs) then (coeffs, i)
    else Polynomial.createPolyInter rest f n);
createPolyInter [] f n = (Polynomial.ones n, Util.intToCint 0);

createRedPairProc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Polynomial.RedPair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             [(Term.Term a b, Term.Term a b)] ->
                               Util.Eithera [[Char]] ();
createRedPairProc interpr p rP r =
  let {
    alpha =
      Polynomial.createPolyInter (Polynomial.redPairToPrePolyInter interpr);
  } in RedPair.checkRedPairProc (Polynomial.checkSPolo alpha)
         (Polynomial.checkNSPolo alpha) "polynomial order"
         (Polynomial.createRepresentation alpha
           (List.remdups
             (List.mapa (\ (a @ (g, il)) -> (g, List.size_list (snd il)))
               (Polynomial.redPairToPrePolyInter interpr))))
         p rP r;

checkCintNonnegative :: Util.Cint -> Util.Eithera [[Char]] ();
checkCintNonnegative (Util.Neg n) =
  ErrorMonad.checkB (Nat.eq_nat n Nat.Zero_nat) ["Negative constant."];
checkCintNonnegative (Util.Pos uu) = ErrorMonad.returna ();

checkNegPoloMono ::
  forall a. [(a, (Util.Cint, [Nat.Nat]))] -> Util.Eithera [[Char]] ();
checkNegPoloMono ((uu, (c, coeffs)) : alpha) =
  ErrorMonad.bind
    (ErrorMonad.bind (Polynomial.checkCintNonnegative c)
      (\ x -> Polynomial.checkListPositive coeffs))
    (\ x -> Polynomial.checkNegPoloMono alpha);
checkNegPoloMono [] = ErrorMonad.returna ();

checkInterMono :: forall a. Polynomial.RedPair a -> Util.Eithera [[Char]] ();
checkInterMono (Polynomial.NegPolo alpha) = Polynomial.checkNegPoloMono alpha;
checkInterMono (Polynomial.Polo alpha) = Polynomial.checkPoloMono alpha;

createMonoRedPairProc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Polynomial.RedPair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             [(Term.Term a b, Term.Term a b)] ->
                               [(Term.Term a b, Term.Term a b)] ->
                                 Util.Eithera [[Char]] ();
createMonoRedPairProc interpr p rP r rR =
  let {
    alpha =
      Polynomial.createPolyInter (Polynomial.redPairToPrePolyInter interpr);
  } in ErrorMonad.bind (Polynomial.checkInterMono interpr)
         (\ x ->
           RedPair.checkMonoRedPairProc (Polynomial.checkSPolo alpha)
             (Polynomial.checkNSPolo alpha) "polynomial order"
             (Polynomial.createRepresentation alpha
               (List.remdups
                 (List.mapa (\ (a @ (g, il)) -> (g, List.size_list (snd il)))
                   (Polynomial.redPairToPrePolyInter interpr))))
             p rP r rR);

createRuleRemovalProc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Polynomial.RedPair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             Util.Eithera [[Char]] ();
createRuleRemovalProc interpr r rR =
  let {
    alpha =
      Polynomial.createPolyInter (Polynomial.redPairToPrePolyInter interpr);
  } in ErrorMonad.bind (Polynomial.checkInterMono interpr)
         (\ x ->
           RedPair.checkRuleRemovalProc (Polynomial.checkSPolo alpha)
             (Polynomial.checkNSPolo alpha) "polynomial order"
             (Polynomial.createRepresentation alpha
               (List.remdups
                 (List.mapa (\ (a @ (g, il)) -> (g, List.size_list (snd il)))
                   (Polynomial.redPairToPrePolyInter interpr))))
             r rR);

}
